import os
import pathlib
import re
from concurrent.futures import ThreadPoolExecutor
from configparser import ConfigParser
from typing import Dict, List, Optional, Tuple, Union

import requests
import yaml
from rich.progress import (
    BarColumn,
    DownloadColumn,
    Progress,
    TaskID,
    TextColumn,
    TransferSpeedColumn,
)

# What west's APIs accept for paths.
#
# Here, os.PathLike objects should return str from their __fspath__
# methods, not bytes. We could try to do something like the approach
# taken in https://github.com/python/mypy/issues/5264 to annotate that
# as os.PathLike[str] if TYPE_CHECKING and plain os.PathLike
# otherwise, but it doesn't seem worth it.
PathType = Union[str, os.PathLike]

# Used to keep track of supported python versions
supported_python_versions = [
    {"major": 3, "minor": 8},
    {"major": 3, "minor": 9},
    {"major": 3, "minor": 10},
]

progress = Progress(
    TextColumn("[bold blue]{task.fields[filename]}"),
    BarColumn(),
    DownloadColumn(),
    "•",
    TransferSpeedColumn(),
)


class WestDirNotFound(RuntimeError):
    """Neither the current directory nor any parent has a west workspace."""


class WestConfigNotFound(RuntimeError):
    """.west/config file does not exist."""


class WestYmlNotFound(RuntimeError):
    """west.yml file does not exist."""


class EastJlinkDeviceLoadError(RuntimeError):
    """Some error happened when trying to collect --device flag for JLink."""


def download_file(task_id: TaskID, url: str, path: str):
    """Download a file from the internet to the given path."""
    file_size = requests.head(url, allow_redirects=True).headers.get(
        "content-length", -1
    )
    response = requests.get(url, stream=True)

    progress.update(task_id, total=int(file_size))

    with open(path, "wb") as f:
        progress.start_task(task_id)
        for chunk in response.iter_content(chunk_size=1024):
            if chunk:
                progress.update(task_id, advance=len(chunk))
                f.write(chunk)


def download_files(files_to_download: List[Dict], dest_dir: str) -> List[str]:
    """Download concurrently multiple files from the internet to the given directory.

    Function expects a list of dicts, where each dict has a key "url" that points to the
    file that should be downloaded and a key "name" that is used for naming and printing
    purposes.

    After all files were downloaded the function returns a list of paths to the
    downloaded files in the same order as they were given.

    Args:
        files_to_download (List[Dict]):     Files to download.
        dest_dir (List[Dict]):              Directory where to download the files.

    Return:
        files (List[str]):      List of paths to the downloaded files.
    """
    file_paths = []

    with progress:
        with ThreadPoolExecutor(max_workers=4) as pool:
            for file in files_to_download:
                filename = file["name"]

                if not os.path.isdir(dest_dir):
                    os.mkdir(dest_dir)

                dest_path = os.path.join(dest_dir, filename)
                file_paths.append(dest_path)

                task_id = progress.add_task("download", filename=filename, start=False)
                pool.submit(download_file, task_id, file["url"], dest_path)

    return file_paths


def west_topdir(start: Optional[PathType] = None) -> str:
    """Returns the path to the parent directory of the .west/
    directory instead, where project repositories are stored.

    Args:
        start (Optional[PathType]):     Directory from where to start searching, if not
                                        given current directory is used.

    Returns:
        Full path to parent directory of the .west/ folder.
    """
    cur_dir = pathlib.Path(start or os.getcwd())

    while True:
        if (cur_dir / ".west").is_dir():
            return os.fspath(cur_dir)

        parent_dir = cur_dir.parent
        if cur_dir == parent_dir:
            # We are at top level
            raise WestDirNotFound(
                "Could not find a west workspace in this or any parent directory"
            )
        cur_dir = parent_dir


def get_ncs_and_project_dir(west_dir_path: str) -> Tuple[str, str]:
    """Returns version of nrf-sdk project and absolute path to the projects directory.

    This is combined, so we avoid reading .west/config file twice.

    Args:
        west_dir_path (str):        Path to the parent of .west directory. It is up to
                                    the caller to provide a correct value.

    Returns:
        Revision string of nrf-sdk project, absolute

    Raises:
        WestConfigNotFound or WestYamlNotFound
    """
    config = ConfigParser()
    config_path = os.path.join(west_dir_path, ".west", "config")

    # ".west/config file could not exist, in that case we should raise an exception"
    if not os.path.isfile(config_path):
        raise WestConfigNotFound(".west/conifg file does not exists.")

    config.read(config_path)

    # Get path to west.yaml file
    project_path = os.path.join(west_dir_path, config["manifest"]["path"])
    west_yaml = os.path.join(project_path, config["manifest"]["file"])

    if not os.path.isfile(west_yaml):
        raise WestYmlNotFound("No west.yml was found ")

    # Get ncs version
    with open(west_yaml, "r") as file:
        manifest = yaml.safe_load(file)["manifest"]

    try:
        ncs = list(
            filter(
                lambda project: project["repo-path"] == "sdk-nrf", manifest["projects"]
            )
        )
    except KeyError:
        # This can happen in the case where there is no sdk-nrf repo in the west yaml
        # file, project is probably using ordinary Zephyr.
        return None, project_path

    return (ncs[0]["revision"], project_path)


def return_dict_on_match(array_of_dicts, key, value):
    """Search through array of dicts and return the first one where the given key matches
    the given value.
    If none are found return None.
    """
    return next((item for item in array_of_dicts if item.get(key) == value), None)


no_toolchain_manager_msg = """
[bold magenta]nrfutil toolchain-manager[/] is [bold red]not installed[/] on this system!

To install it run:

\t[italic bold blue]east install nrfutil-toolchain-manager
"""

not_in_west_workspace_msg = """
[bold yellow]West workspace[/] was [bold red]not found![/]

This command can only be run [bold]inside[/] of a [bold yellow]West workspace[/].
"""


def no_toolchain_msg(east):
    """Return message that informs user that toolchain is not installed."""
    return (
        f"Current [bold cyan]NCS[/] [bold]{east.detected_ncs_version}[/] version is "
        "supported but toolchain is [bold red]not installed![/]"
        "\n\nTo install it run:"
        "\n\n\t[italic bold blue]east install toolchain\n"
    )


def ncs_version_not_supported_msg(east, supported_versions):
    """Return message that informs user that detected NCS version is not supported."""
    vers = "\n".join(
        [f"[bold yellow]•[/] {ver}" for ver in supported_versions.strip().split("\n")]
    )

    return (
        f"[bold]East[/] detected [bold]{east.detected_ncs_version}[/] [bold cyan]NCS[/]"
        " version which is currently [bold red]not supported[/] by the "
        "[bold magenta]nrfutil toolchain-manager[/].\n\nSupported versions are: \n"
        + vers
        + "\n\nThis means that you need to manually install the [bold cyan]NCS[/]"
        " toolchain by yourself.\n"
    )


BOARD_YML = "board.yml"


def find_board_dir(east, board: str):
    """Find the directory of the given board.

    Search for directory that contains *_defconfig file with given west_board name,
    (this is the same process that west uses to determine the boards).

    After such directory is found, scan it and try to find what board revisions are
    present and generate a list of board names and board revisions in format expected by
    west build command.
    """
    root = os.path.join(east.project_dir, "boards")

    for path, _, files in os.walk(root):
        # Presence of board.yml file is used to detect hw v2 boards
        if BOARD_YML in files:
            with open(os.path.join(path, BOARD_YML)) as f:
                if yaml.safe_load(f).get("board").get("name") in board:
                    return path

        # Fallback to the old method of searching
        for file in files:
            if file.endswith("_defconfig") and file[: -len("_defconfig")] in board:
                return path

    return None


def find_all_boards(east, west_board: str) -> List[str]:
    """Find all possible revisions of west board.

    If there are no revision specific files in the folder then just return west_board.

        east ():            East context.
        west_board (str):   Board to search for.

    Returns:
        List of west boards to be used by west build -b <board>
    """
    west_board_unmodified = west_board

    # "Normalize" hw v2 board names
    if "/" in west_board:
        west_board = west_board.replace("/", "_")

    board_dir = find_board_dir(east, west_board)

    if not board_dir:
        # Just return the original name, since the board might come from Zephyr or NCS.
        return [west_board]

    files = os.listdir(board_dir)

    # Find all files with west board name, no matter the version
    pattern = rf"^{west_board}_[0-9]?[0-9]?_[0-9]?[0-9]?_[0-9]?[0-9]?\.conf"
    matches = [file for file in files if re.match(pattern, file)]

    # Extract hardware versions and put them into format expected by the west.
    hw_versions = [
        ".".join(m.split(".")[0].replace(west_board, "").split("_")[1:])
        for m in matches
    ]
    boards = ["@".join([west_board_unmodified, hw]) for hw in sorted(hw_versions)]

    return boards if boards else [west_board_unmodified]


def clean_up_extra_args(args):
    """Clean up extra args, by adding back double quotes to the define assignments.

    Click argument automatically strips double quotes from anything that is given
    after "--". We can not know for sure from where double quotes were removed,
    however we know that CMake in at least one case requires them.
    For example, double quotes are needed, if you are assigning a list of tokens to the
    define value, like so: -DFILES="file1.txt;file2.txt".
    Double quotes are also added if there is an assignment to a flag, like:
    --some-flag=something.
    If this line gets to Cmake without double quotes, it will not be parsed
    correctly.

    We also take care of the case where there is a space in the argument, in that case
    we add double quotes around the whole argument.

        args ():    Extra args, passed after '--' as Click argument.

    Returns:
        cleaned up args, already formatted as a string.
    """

    def add_back_double_quotes(arg):
        if arg.startswith("-") and "=" in arg:
            # Split by the first "=" and add double quotes to the value
            split = arg.split("=", 1)
            arg = f'{split[0]}="{split[1]}"'
        else:
            # The arg is not an option of any kind, just add double quotes if it
            # contains a space
            if " " in arg:
                arg = f'"{arg}"'
        return arg

    return f"{' '.join(list(map(add_back_double_quotes, args)))}"


def create_artefact_name(project, board, version, build_type):
    """Create an artefact name.

    Board might be in form <west_board>@<hv_version>, in that case we modify it to fit
    the artefact name.

    We also add git hash at the end if the build was not done on the clean tagged
    commit.
    """
    board = board.replace("@", "-hv")

    # "release" or None build_type should not generate any build type qualifier.
    build_type = "" if build_type == "release" or not build_type else f"-{build_type}"

    git_hash = f"-{version['hash']}" if version["hash"] else ""

    return f"{project}-{board}-{version['tag']}{build_type}{git_hash}"


def get_git_version(east):
    """Return structured output from git describe command.

    Note: See help string of release function for more information.

    Note to myself: If you use this function in some other place, make sure that you
    patch it in the test files, otherwise pytest thinks that is the mocked call that it
    should test against.
    """
    result = east.run(
        "git describe --tags --always --long --dirty=+", silent=True, return_output=True
    )

    output = result["output"].strip().split("-")

    if len(output) == 1:
        # No git tag, only hash was produced
        return {"tag": "v0.0.0", "hash": output[0]}

    if len(output) >= 3:
        # Some tags have '-' in them, like the ncs ones (eg. v3.3.99-ncs1)
        version = {"tag": "-".join(output[:-2])}

        if output[-2] == "0" and not output[-1].endswith("+"):
            # Clean version commit, no hash needed
            version["hash"] = ""
        else:
            # Not on commit or dirty, both version and hash are needed
            version["hash"] = output[-1][1:]

        return version

    east.print(
        f"Unsupported git describe output ({result['output']}), contact developer!"
    )
    east.exit()


def get_device(runner_yaml_content):
    """Extract device flag from runner.yaml content."""
    try:
        jlink_args = runner_yaml_content["args"]["jlink"]
        device = next(filter(lambda e: "--device" in e, jlink_args))
        return device.split("=")[1]
    except (KeyError, StopIteration):
        return None


def find_app_build_dir(build_dir):
    """Find the build directory of the app."""
    try:
        # Presence of domain.yaml indicates that the project is using sysbuild.
        with open(os.path.join(build_dir, "domains.yaml")) as f:
            d = yaml.safe_load(f)
            return os.path.join(build_dir, d["default"], "zephyr")
    except FileNotFoundError:
        return os.path.join(build_dir, "zephyr")


def get_device_in_runner_yaml(build_dir):
    """Returns device flag for JLink runner from runners.yaml.

    If it is not possible to do that it will raise an exception.
    """
    app_build_dir = find_app_build_dir(build_dir)

    with open(os.path.join(app_build_dir, "runners.yaml")) as f:
        runners_yaml = yaml.safe_load(f)

    device = get_device(runners_yaml)

    if not device:
        raise Exception(
            "No JLink device was found in the runners.yaml file! Used board might "
            "not be supported. Try specifying the specific device with --device flag."
        )

    return device


def get_cortex_debug_params(build_dir):
    """Get cortex debug parameters from the build directory.

    Any path returned will be absolute.

    Any failure to read something will result in a raised exception so this should be
    called inside the try block.
    """
    if not os.path.isdir(build_dir):
        raise Exception(f"Build directory {build_dir} not found")

    runners_file = os.path.join(find_app_build_dir(build_dir), "runners.yaml")
    elf_file = os.path.join(find_app_build_dir(build_dir), "zephyr.elf")
    elf_file = os.path.realpath(elf_file)

    with open(runners_file) as f:
        runner_cfg = yaml.safe_load(f)

    device = get_device(runner_cfg)

    if not device:
        raise Exception(f"Can't find JLink device in the {runners_file} file")

    if "gdb" not in runner_cfg["config"]:
        raise Exception(f"Can't find gdb path in the {runners_file} file")

    if not os.path.isfile(elf_file):
        raise Exception(f"Elf file {elf_file} not found")

    return device, runner_cfg["config"]["gdb"], elf_file


def determine_svd_file(east, device):
    """Determine the SVD file for the given device.

    Only Nordic devices are supported at the moment, unsupported devices will return
    None.
    """
    mdk_dir = os.path.join(
        east.west_dir_path,
        "modules",
        "hal",
        "nordic",
        "nrfx",
        "mdk",
    )

    if device.startswith("nRF5340"):
        if device.endwith("NET"):
            svd = "nrf5340_network"
        elif device.endwith("APP"):
            svd = "nrf5340_application"
        else:
            return None
    elif device.startswith("nRF52832"):
        svd = "nrf52"
    elif device.startswith("nRF51"):
        svd = "nrf51"
    else:
        svd = device.split("_")[0].lower()

    return next(
        (
            os.path.join(mdk_dir, file)
            for file in os.listdir(mdk_dir)
            if svd == os.path.splitext(file)[0] and file.endswith(".svd")
        ),
        None,
    )



import click

from ..east_context import east_command_settings, east_group_settings
from .tooling import tool_installer


@click.command(**east_command_settings)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    help="Reinstall the toolchain, even if it is already installed.",
)
@click.pass_obj
def toolchain(east, force):
    """Install NCS toolchain for the current West workspace.

    \b
    \n\nDetermines the version of the NCS that is being used in the current West
    workspace and downloads the correct version of the toolchain for it.

    \n\n[bold]Note:[/] This command can be only run from inside of a [bold yellow]West workspace[/].
    """
    east.pre_workspace_command_check(
        ignore_uninstalled_ncs=True,
        ignore_unsupported_ncs=False,
    )

    # # We come to here if the detected ncs version is installed.
    if east.detected_ncs_version_installed:
        if not force:
            msg = (
                f"\nVersion of [bold cyan]{east.detected_ncs_version} NCS[/] toolchain "
                "is [bold]already[/] installed!\n\n"
                "You can reinstall it with [bold cyan]--force[/] flag. "
            )
            east.print(msg)

            # Exit with 0 error code, we do that so it makes usage of East in CI easier.
            east.exit(0)

        east.run_manager(f"uninstall --ncs-version {east.detected_ncs_version}")
        # Empty line for nicer output
        east.print()

    east.print(
        "Starting install of [bold cyan]NCS[/] toolchain,"
        f" version [bold]{east.detected_ncs_version}[/], this will take some time...\n",
        highlight=False,
    )
    east.run_manager(f"install --ncs-version {east.detected_ncs_version}")

    east.print("\n[bold green]Done!")


@click.command(**east_command_settings)
@click.pass_obj
def nrfutil_toolchain_manager(east):
    """Install [bold magenta]nrfutil toolchain-manager[/]."""
    tool_installer(east, ["toolchain-manager"])


@click.command(**east_command_settings)
@click.pass_obj
def codechecker(east):
    """Install [bold magenta]CodeChecker[/] and its dependencies.

    \b
    \n\nCurrent dependencies are [bold cyan]clang-tidy[/] and [bold cyan]clang[/], both are contained in the [bold cyan]clang+llvm[/] package.
    """
    tool_installer(east, ["codechecker", "clang+llvm", "cppcheck"])


@click.option(
    "--all",
    is_flag=True,
    help="Install all packages.",
)
@click.group(
    **east_group_settings,
    subcommand_metavar="package_name",
    invoke_without_command=True,
)
@click.pass_obj
@click.pass_context
def install(ctx, east, all):
    """Install one or more tools.

    \b
    \n\nTools are always installed in the [bold magenta]~/.local/share/east/tooling[/] directory.

    \n\nTo install a single tool run [bold cyan]east install tool_name[/].
    \n\nTo install all tools run [bold cyan]east install --all[/].
    \n\nTo learn more about a tool run [bold cyan]east install tool_name --help[/].

    """
    # Since we want to call commands in code and not through cli, we need to use
    # .callback method which is actually the original funciton without click's
    # infrastructure.
    # WARN: There is a better way to do this: invoke method

    if all:
        tool_installer(east, ["codechecker", "clang+llvm", "toolchain-manager"])
        ctx.invoke(toolchain)
    return


install.add_command(toolchain)
install.add_command(codechecker)
install.add_command(nrfutil_toolchain_manager)



import glob
import json
import linecache
import os
import plistlib as plist
import re

from ..helper_functions import get_git_version


def check_for_codechecker(east):
    """Check if codechecker is installed and available in the path."""
    if not east.check_exe(east.consts["codechecker_path"]):
        msg = (
            "\n[bold magenta]Codechecker[/] is [bold red]not installed[/] on this "
            "system!\n\n"
            "To install it run:\n\n"
            "\t[italic bold blue] east install codechecker[/]"
        )
        east.print(msg)
        east.exit()


def check_for_build_folder(east, build_dir):
    """Check that build dir exists."""
    if not os.path.isdir(build_dir):
        east.print(
            f"\nBuild folder [magenta bold italic]{build_dir}[/] was [bold red]not found[/].\n\n"
            "Build your project first and rerun this command."
        )
        east.exit()


def check_for_compile_commands_json(east, compile_commands):
    """Check that compile_commands.json exists in the build dir."""
    if not os.path.exists(compile_commands):
        east.print(
            f"\nFile [cyan bold]{compile_commands}[/] was [bold red]not found[/].\n"
            "Check if your project's [cyan bold]CMakelists.txt[/] file contains the below line:\n\n"
            "[bold]set(CMAKE_EXPORT_COMPILE_COMMANDS ON)[/]\n"
        )
        east.exit()


def cleanup_compile_commands_json(compile_commands):
    """Remove GCC specific arguments from compile_commands.json.

    Otherwise clangsa fails to analyze the files.
    """
    args = [
        "-fno-reorder-functions",
        "-fno-freestanding",
        "-mfp16-format=ieee",
        "-fno-printf-return-value",
    ]

    # Create matching regex for all arguments
    args_regex = "|".join([f"({arg})" for arg in args])

    with open(compile_commands, "r") as f:
        data = json.load(f)

    for entry in data:
        entry["command"] = re.sub(f"{args_regex}", "", entry["command"])

    with open(compile_commands, "w") as f:
        f.write(json.dumps(data))


def check_for_codechecker_config_yaml(east, cfg):
    """Check that codechecker_config.yaml exists in the project's root dir."""
    if not os.path.exists(cfg):
        east.print(
            "\nFile [cyan bold]codechecker_config.yaml[/] was [bold red]not found[/] in project's git root dir.\n"
            "CodeChecker requires this file to function normally.\n\n"
            "Run below command to see an example of a config: \n\n"
            "\t[cyan bold]east codechecker example-config\n"
        )
        east.exit()


def detect_problematic_zephyr_macros(diag):
    """Return true if a problematic Zephyr macros is flagged by clang-tidy diagnostics.

    Macros are always connected to a specific description.
    Different problematic macros can have the same description.
    """
    bad_diagnostics = [
        {
            "description": "ineffective bitwise and operation",
            "patterns": ["expanded from macro '.*LOG_.*'"],
        },
        {
            "description": "conditional operator with identical true and false expressions",
            "patterns": [
                "expanded from macro 'SHELL_CMD_.*'",
                "expanded from macro 'APP_EVENT_TYPE_DEFINE'",
            ],
        },
        {
            "description": "comparison of integers of different signs: 'int' and 'unsigned int'",
            "patterns": ["expanded from macro 'INIT_OBJ_RES.*'"],
        },
        {
            "description": "missing field 'help' initializer",
            "patterns": ["expanded from macro 'SHELL_SUBCMD_.*'"],
        },
    ]

    # Iterate through all bad diagnostics and first check if the descriptions match.
    # If yes, then traverse the paths and check if any of the patterns appear in the
    # "message" field.
    for bad_diagnostic in bad_diagnostics:
        if diag["description"] == bad_diagnostic["description"]:
            for path in diag["path"]:
                if "message" in path:
                    for pattern in bad_diagnostic["patterns"]:
                        if re.search(pattern, path["message"]):
                            return True
    return False


def var_never_read_before_disabled_macro_found(files, diag):
    """Returns true if clang-tidy diagnostic is about value being stored to a variable,
    and not being read, but this variable is later passed into some kind of a disabled
    macro, such as __ASSERT* or LOG_* macro.
    """
    # Determine first if the description of the diagnostic is as expected for this case
    desc_regexes = [
        "Value stored to '(.*)' during its initialization is never read",
        "Value stored to '(.*)' is never read",
        "unused variable '(.*)'",
    ]

    var = None
    for desc_regex in desc_regexes:
        hit = re.search(desc_regex, diag["description"])
        if hit:
            # Extracted variable name
            var = hit.group(1)
            break

    if not var:
        return False

    # A hit was found, locate the file where the diagnostic was found
    file = files[diag["location"]["file"]]

    # Open it and start reading from the line where the diagnostic was found
    with open(file, "r") as f:
        for _ in range(diag["location"]["line"]):
            next(f)

        for line in f:
            # If you hit a line where variable is passed to __ASSERT* or LOG_* macro,
            # then this is a false positive, and should be removed
            patterns = [
                rf".*__ASSERT.*\(.*{var}.*,.*",
                rf".*LOG_.*\(.*,.*{var}.*",
            ]

            for pattern in patterns:
                if re.search(pattern, line):
                    return True

            # If you hit a line that uses the variable in any way, then this is a true
            # positive
            if re.search(f".*{var}.*", line):
                return False


def sizeof_on_pointer_type_found(files, diag):
    """Returns true if clang-tidy diagnostic is about sizeof() being called on a
    pointer, inside a LOG_*, EVENT_LOG* or APP_EVENT_MANAGER_LOG macro.
    """
    # Determine first if the description of the diagnostic is as expected for this case
    desc = "The code calls sizeof() on a pointer type. This can produce an unexpected result"

    if diag["description"] != desc:
        return False

    # A hit was found, locate the file and the line where the diagnostic was found
    line = linecache.getline(files[diag["location"]["file"]], diag["location"]["line"])

    # If you hit a line where variable is passed to __ASSERT* or LOG_* macro,
    # then this is a false positive, and should be removed
    patterns = [
        ".*LOG_.*",
        ".*EVENT_LOG.*",
        ".*APP_EVENT_MANAGER_LOG.*",
    ]

    for pattern in patterns:
        if re.search(pattern, line):
            return True

    return False


def cleanup_plist_files(east, output):
    """Cleanup generated plist files by removing diagnostics that are not useful."""
    # Conveniance variable, as the listeral is quite long
    hash = "issue_hash_content_of_line_in_context"

    for file in glob.glob(os.path.join(output, "*.plist")):
        with open(file, "rb") as f:
            data = plist.load(f)

        diags = data["diagnostics"]

        diags_for_removal = []

        # Iterate through all diagnostics in the plist file and filter ones that you do
        # not want.
        # You can place any custom logic below to filter out diagnostics that you
        # do not want.
        for diag in diags:
            if (
                detect_problematic_zephyr_macros(diag)
                or var_never_read_before_disabled_macro_found(data["files"], diag)
                or sizeof_on_pointer_type_found(data["files"], diag)
            ):
                diags_for_removal.append(diag[hash])

        # Removal happens here
        data["diagnostics"] = [
            diag for diag in diags if diag[hash] not in diags_for_removal
        ]

        # Write the plist file back
        with open(file, "wb") as f:
            plist.dump(data, f)


def create_skip_file(east, build_dir, output):
    """Create skip file for CodeChecker.

    Generated skip file skips analysis of Zephyr, NCS, external repositories and
    contents of any build folder inside the project's root dir.

    We are skipping build folder because it contains generated .c files that are not in
    other folders.
    """
    skip_file = os.path.join(output, "skip_file.txt")

    os.makedirs(output, exist_ok=True)

    filename = os.path.join(build_dir, "codecheckerfile.json")

    # Read the build_dir from the codecheckerfile.json file
    with open(filename, "r") as f:
        build_dir = json.load(f)["build_dir"]

    if not os.path.exists(skip_file):
        with open(skip_file, "w") as f:
            f.write(f"-{build_dir}/*\n")
            f.write(f"+{east.project_dir}/*\n")
            f.write(f"-{east.west_dir_path}/*\n ")

    return skip_file


def create_codecheckerfile(east, board, build_type, build_dir, source_dir):
    """Create a file in the build folder that contains all information needed for the
    codechecker store command.

    The intention is that east codechecker store reads the file and uses it for
    creating the metadata of the run when storing to the server.
    """
    # Create a file in the build folder that contains the name of the artefact
    name = os.path.basename(source_dir if source_dir else east.cwd)
    build_dir = os.path.join(east.cwd, build_dir if build_dir else "build")
    source_dir = os.path.join(east.cwd, source_dir if source_dir else "")
    filename = os.path.join(build_dir, "codecheckerfile.json")
    version = get_git_version(east)

    data = {
        "name": name,
        "board": board,
        "build_type": build_type,
        "version": version,
        "build_dir": build_dir,
        "source_dir": source_dir,
    }
    if not os.path.exists(build_dir):
        os.makedirs(build_dir)

    with open(filename, "w") as f:
        json.dump(data, f)


def get_metadata_from_codecheckerfile(build_dir):
    """Read the codecheckerfile.json file and construct the name and the tag, intended
    to be used by codechecker store command.
    """
    filename = os.path.join(build_dir, "codecheckerfile.json")
    with open(filename, "r") as f:
        data = json.load(f)

    build_type = data["build_type"]
    build_type = "" if build_type == "release" or not build_type else f"-{build_type}"

    name = f"{data['name']}-{data['board']}{build_type}"

    tag = f"tag: {data['version']['tag']}"

    if data["version"]["hash"]:
        tag += f", commit: {data['version']['hash']}"

    return name, tag


def get_endpoint(east):
    """Get the endpoint from the project's git remote origin."""
    cmd = "git config --get remote.origin.url"
    result = east.run(cmd, return_output=True, silent=True)
    return result["output"].split("/")[-1].split(".")[0].strip()



import click

from ..east_context import east_command_settings
from ..helper_functions import clean_up_extra_args


@click.command(
    **east_command_settings,
    context_settings=dict(ignore_unknown_options=True, allow_extra_args=True),
)
@click.pass_obj
@click.option(
    "--extra-help",
    is_flag=True,
    help="Print help of the [bold magenta]west init[/] command.",
)
@click.argument("args", nargs=-1, type=click.UNPROCESSED, metavar="")
def init(east, extra_help, args):
    """Create a [bold yellow]West workspace[/].

    \b
    \n\nInternally runs [magenta bold]west init[/] command, all given arguments are passed directly to it.

    \n\nTo learn more about possible [magenta bold]west init[/] arguments and options use --extra-help flag.


    \n\n[bold]Note:[/] This command can be only run from inside of a [bold yellow]West workspace[/].
    """
    cmd = "init "

    if extra_help:
        cmd += "--help"
        east.run_west(cmd)
        east.exit(return_code=0)

    if args:
        cmd += f"{clean_up_extra_args(args)} "

    east.run_west(cmd)


@click.command(
    **east_command_settings,
    context_settings=dict(ignore_unknown_options=True, allow_extra_args=True),
)
@click.pass_obj
@click.option(
    "--extra-help",
    is_flag=True,
    help="Print help of the [bold magenta]west update[/] command.",
)
@click.argument("args", nargs=-1, type=click.UNPROCESSED, metavar="")
def update(east, extra_help, args):
    """Update a [bold yellow]West workspace[/].

    \b
    \n\nInternally runs [magenta bold]west update[/] command, all given arguments are passed directly to it.

    \n\nTo learn more about possible [magenta bold]west update[/] arguments and options use --extra-help flag.


    \n\n[bold]Note:[/] This command can be only run from inside of a [bold yellow]West workspace[/].
    """
    cmd = "update "

    if extra_help:
        cmd += "--help"
        east.run_west(cmd)
        east.exit(return_code=0)

    if args:
        cmd += f"{clean_up_extra_args(args)} "

    east.run_west(cmd)



import os

# From this file only dict consts_path should be imported, and this should happen only
# in __main__.
# There are however exceptions: NRFUTIL_PATH, CPPCHECK_PATH,
# CLANG_PATH, CODECHECKER_PATH are needed for downloading in install.py.

HOME_DIR = os.environ["HOME"]

# Directory that can be used for temporary intermediate files that we do not care about.
CACHE_DIR = os.path.join(HOME_DIR, ".cache", "east")

# East directory, any files that east needs to function normally are located here.
EAST_DIR = os.path.join(HOME_DIR, ".local", "share", "east")

# Tooling directory, for all the tools that east uses.
TOOLING_DIR = os.path.join(EAST_DIR, "tooling")

# Path to the nrfutil executable.
NRFUTIL_PATH = os.path.join(TOOLING_DIR, "nrfutil", "nrfutil")

# Path to the cppcheck executable.
CPPCHECK_PATH = os.path.join(TOOLING_DIR, "cppcheck", "cppcheck")

# Path to the clang executables.
CLANG_BIN_PATH = os.path.join(TOOLING_DIR, "clang+llvm", "bin")
CLANG_PATH = os.path.join(CLANG_BIN_PATH, "clang")
CLANG_TIDY_PATH = os.path.join(CLANG_BIN_PATH, "clang-tidy")
CLANG_REPLACE_PATH = os.path.join(CLANG_BIN_PATH, "clang-apply-replacements")

# Path to the CodeChecker executable.
CODECHECKER_PATH = os.path.join(
    TOOLING_DIR, "codechecker", "build", "CodeChecker", "bin", "CodeChecker"
)

if os.environ.get("EAST_CODECHECKER_CI_MODE", "0") == "1":
    # In CI mode, the CodeChecker executable is not in the tooling directory, but in
    # the system path.
    CODECHECKER_PATH = "CodeChecker"

# Directory will all Conda stuff.
MINICONDA_DIR = os.path.join(HOME_DIR, "miniconda3")

# Path to the Conda executable, this can be used when the Conda is not yet on PATH.
CONDA_PATH = os.path.join(MINICONDA_DIR, "bin", "conda")

const_paths = {
    "cache_dir": CACHE_DIR,
    "east_dir": EAST_DIR,
    "tooling_dir": TOOLING_DIR,
    "nrfutil_path": NRFUTIL_PATH,
    "miniconda_dir": MINICONDA_DIR,
    "conda_path": CONDA_PATH,
    "cppcheck_path": CPPCHECK_PATH,
    "clang_path": CLANG_PATH,
    "clang_tidy_path": CLANG_TIDY_PATH,
    "clang_replace_path": CLANG_REPLACE_PATH,
    "codechecker_path": CODECHECKER_PATH,
}



import os

from east.east_context import EastContext

from . import helpers


def assert_all_is_none(east, path):
    """Small helper for brevity."""
    assert east.cwd == path
    assert east.west_dir_path is None
    assert east.detected_ncs_version is None
    assert east.project_dir is None


def test_good_west_workplace(west_workplace_parametrized):
    """Test a good west workplace."""
    project_path = west_workplace_parametrized["project"]

    east = EastContext()
    assert east.cwd == project_path
    assert east.west_dir_path == os.path.dirname(project_path)
    assert east.detected_ncs_version == "v2.1.0"

    assert east.use_toolchain_manager is False
    assert east.east_yml is None

    east.pre_workspace_command_check()
    assert east.use_toolchain_manager is True
    assert east.east_yml is not None


def test_no_config_west_workplace(west_workplace_parametrized):
    """Test a west workplace without a .west/config file."""
    project_path = west_workplace_parametrized["project"]
    os.remove(os.path.join(os.path.dirname(project_path), ".west", "config"))

    east = EastContext()
    assert_all_is_none(east, project_path)


def test_not_in_west_workplace(not_in_west_workplace):
    """Test a path that is not in a west workplace."""
    east = EastContext()
    assert_all_is_none(east, not_in_west_workplace)


def test_not_ncs_sdk_west_workplace(west_workplace_parametrized):
    """Test a west workplace without the NCS SDK."""
    project_path = west_workplace_parametrized["project"]

    helpers.west_no_nrf_sdk_in_yaml(os.path.dirname(project_path))

    east = EastContext()
    assert east.cwd == project_path
    assert east.west_dir_path == os.path.dirname(project_path)
    assert east.detected_ncs_version is None


def test_no_west_yaml_west_workplace(west_workplace_parametrized):
    """Test a west workplace without a west.yml file."""
    project_path = west_workplace_parametrized["project"]

    os.remove(os.path.join(project_path, "west.yml"))

    east = EastContext()
    assert_all_is_none(east, project_path)



import os

from click.testing import CliRunner

from east.__main__ import cli


def test_clean_command(west_workplace_parametrized):
    """Test the clean command."""
    _ = west_workplace_parametrized

    runner = CliRunner()

    os.mkdir("build")
    assert "build" in os.listdir()

    result = runner.invoke(cli, ["clean"])

    assert result.exit_code == 0
    assert "build" not in os.listdir()



import click

from ..east_context import east_command_settings
from ..helper_functions import get_device_in_runner_yaml


def no_jlink_tool_msg(tool):
    """Return a message that a required JLink tool was not found."""
    return (
        f"The required program [bold cyan]{tool}[/] was [bold red]not found,[/]"
        " exiting!"
    )


def fmt_runner_error_msg(exception_msg):
    """Format error message for EastJlinkDeviceLoadError."""
    return (
        "An [bold red]error[/] occurred when trying to extract [bold cyan]--device[/] "
        "flag from [bold yellow]runners.yaml[/] file!\n\n"
        f"[italic yellow]{exception_msg}[/]\n"
    )


@click.command(**east_command_settings)
@click.option(
    "-d",
    "--build-dir",
    type=str,
    default="build",
    help=("Build directory of the project. Default: build."),
)
@click.option(
    "-d",
    "--device",
    type=str,
    help=(
        "Set the target device, required by JLinkExe, i.e. [bold]NRF52840_xxAA[/]. "
        "If not given, east tries to infer the device by looking into --build-dir."
    ),
)
@click.option(
    "-i",
    "--dev-id",
    type=int,
    help=(
        "Identification number of a JLink programmer that should be used for connecting."
    ),
)
@click.option(
    "-p",
    "--rtt-port",
    type=int,
    default=19021,
    help=("Set the RTT Telnet port. Default: 19021."),
)
@click.option(
    "-s",
    "--speed",
    type=str,
    default=4000,
    help=(
        "Set the connection speed, can be a number, 'auto' or 'adaptive'. "
        "Default: '4000'."
    ),
)
@click.pass_obj
def connect(east, device, dev_id, rtt_port, speed, build_dir):
    """Connect to a device and create a RTT server with [bold cyan]JLinkExe[/].

    \b
    \n\nRTT server will emit any RTT messages produced by the device over its dedicated port. Execute [bold magenta]east util rtt[/] command in the separate window to observe these messages.

    """
    if not east.check_exe("JLinkExe"):
        east.print(no_jlink_tool_msg("JLinkExe"), highlight=False)
        east.exit()

    cmd = f"JLinkExe -AutoConnect 1 -Speed {speed} -If SWD -RTTTelnetPort {rtt_port} "

    if dev_id:
        cmd += f"-USB {dev_id} "

    if not device:
        try:
            device = get_device_in_runner_yaml(build_dir)
        except Exception as msg:
            east.print(fmt_runner_error_msg(msg), highlight=False)
            east.exit()

    if device:
        cmd += f"-Device {device} "

    east.run(cmd)


@click.command(**east_command_settings)
@click.option(
    "-e",
    "--local-echo",
    is_flag=True,
    help="Turn on local echo.",
)
@click.option(
    "-p",
    "--rtt-port",
    type=int,
    default=19021,
    help=("Set the RTT Telnet port. Default: 19021."),
)
@click.option(
    "-l",
    "--logfile",
    type=str,
    help="Relative path to a log file into which to save RTT output.",
)
@click.option(
    "-a",
    "--append",
    type=str,
    help="Append RTT output to a log file, instead of overwriting it. Requires --logfile option.",
)
@click.pass_obj
def rtt(east, local_echo, rtt_port, logfile, append):
    """Run a RTT client and connect a running RTT server.

    \b
    \n\nAny messages that RTT server creates are printed. RTT server can be created with an [bold magenta]east util connect[/] command.
    """
    if not east.check_exe("JLinkRTTClient"):
        east.print(no_jlink_tool_msg("JLinkRTTClient"), highlight=False)
        east.exit()

    local_echo = "On" if local_echo else "Off"

    rtt_cmd = f"JLinkRTTClient -LocalEcho {local_echo} -RTTTelnetPort {rtt_port} "

    if logfile and append:
        rtt_cmd += f"ansi2txt | tee -a {logfile}"
    elif logfile:
        rtt_cmd += f"ansi2txt | tee {logfile}"
    elif append:
        east.print(
            "Cannot use [bold cyan]--append[/] flag without "
            "[bold cyan]--logile[/] flag."
        )
        east.exit()

    east.run(rtt_cmd)



import importlib.metadata
import inspect
import os
import signal
import subprocess
import sys
import time
from shutil import which

import requests
from rich.console import Console
from rich.markdown import Markdown
from rich.panel import Panel
from rich_click import RichCommand, RichGroup

from .constants import const_paths
from .east_yml import EastYmlLoadError, format_east_yml_load_error_msg, load_east_yml
from .helper_functions import (
    WestConfigNotFound,
    WestDirNotFound,
    WestYmlNotFound,
    get_ncs_and_project_dir,
    ncs_version_not_supported_msg,
    no_toolchain_manager_msg,
    no_toolchain_msg,
    not_in_west_workspace_msg,
    west_topdir,
)

# Needs to be exposed like this so it can be set to False in tests.
RICH_CONSOLE_ENABLE_MARKUP = True

"""
Convenience dicts for storing settings that are identical across Click's commands and
groups.
"""
east_command_settings = {
    "cls": RichCommand,
    "options_metavar": "[options]",
}

east_group_settings = {
    "cls": RichGroup,
    "options_metavar": "[options]",
}


class EastContext:
    """Context-aware API wrapper & state-passing object.

    EastContext object needs to be created in the top level cli command that groups all
    other commands. That object then needs to be passed to every other subcommand.

    Specifically, the class offers wrappers for core API calls (such as `.run`)
    which take into account CLI parser flags, configuration files, and/or
    changes made at runtime.
    """

    def __init__(self, echo: bool = False):
        """Create a new context object.

        Args:
            echo (bool): If True `.run` prints the command string to local stdout prior
            to executing it. Default: ``False``.
        """
        # This init will be called on true command invocation, --help flag or similar
        # do not count.
        self.cwd = os.getcwd()
        self.echo = echo
        self.consts = const_paths

        # Create east, tooling and cache dirs, if they do not exist.
        os.makedirs(self.consts["east_dir"], exist_ok=True)
        os.makedirs(self.consts["tooling_dir"], exist_ok=True)
        os.makedirs(self.consts["cache_dir"], exist_ok=True)

        self.console = Console(width=80, markup=RICH_CONSOLE_ENABLE_MARKUP)
        self.use_toolchain_manager = False
        self.detected_ncs_version_installed = False
        self.east_yml = None
        self.in_docker = os.path.exists("/.dockerenv")

        try:
            self.west_dir_path = west_topdir()
            self.detected_ncs_version, self.project_dir = get_ncs_and_project_dir(
                self.west_dir_path
            )

        except (WestDirNotFound, WestConfigNotFound, WestYmlNotFound):
            self.west_dir_path = None
            self.detected_ncs_version = None
            self.project_dir = None

        self.check_for_new_east_version()

    def chdir(self, path: str):
        """Change directory.

        self ():
        path (str): Relative or absolute path of the directory to change to.
        """
        os.chdir(path)
        self.cwd = os.getcwd()

    def print(self, *objects, **kwargs):
        """Prints to the console.

        Internally it uses Console object, so whatever Console can do, this function can
        also do.

        Full documentation here:
        https://rich.readthedocs.io/en/latest/reference/console.html#rich.console.Console.print
        """
        self.console.print(*objects, **kwargs)

    def print_info(self, info: str):
        """Prints a message, with many rich settings disabled, with magnify icon
        prepended. Suitable for printing info messages that should not be formatted.
        """
        self.print(
            ":mag_right: " + info,
            markup=True,
            style="bold italic dim",
            overflow="ignore",
            crop=False,
            highlight=False,
            soft_wrap=False,
            no_wrap=True,
        )

    def print_markdown(self, *objects, **kwargs):
        """Interprets given object (string) as Markdown style text and prints it to the
        console.

        Internally it uses Markdown object, so whatever Markdown can do, this function
        can also do.
        https://rich.readthedocs.io/en/stable/reference/markdown.html#rich.markdown.Markdown

        Bonus thing: Any kwargs that are passed to it are correctly sorted and either
        passed to the Markdown object or to the internal self.print function which uses
        Console object.
        """
        markdown_kwargs = {}
        print_kwargs = {}

        for key, value in kwargs.items():
            # The same key can be in both functions so we have to check against both of
            # them.
            if key in inspect.signature(Markdown).parameters.keys():
                markdown_kwargs[key] = value
            if key in inspect.signature(self.console.print).parameters.keys():
                print_kwargs[key] = value

        self.print(Markdown(*objects, **markdown_kwargs), **print_kwargs)

    def exit(self, return_code: int = 1):
        """Exit program with given return_code."""
        sys.exit(return_code)

    def run(
        self,
        command: str,
        exit_on_error: bool = True,
        return_output: bool = False,
        silent: bool = False,
        ignore_sigint: bool = False,
    ) -> dict:
        """Executes given command in shell as a process.

        This is a blocking call, process needs to finish before this command can return;

        Args:
            command (str):  Command to execute.

            exit_on_error (str):    If true the program is exited if the return code of
                                    the ran command is not 0.

            return_output (bool):   Return stdout. Note that this will mean that there
                                    might be no colorcodes in the terminal output and
                                    no strerr, due to piping.

            silent (bool):          Do not print command's output.

            ignore_sigint (bool):   If true it does not pass SIGINT to the run process.
                                    This means that the run process will handle SIGINT
                                    by itself. This is useful for running gdb.

        Returns:
            Dict with two keys is always returned:
                output(str):        Contains stdout of the process that run, if
                                    return_output is True, otherwise empty string.
                returncode(int):    Return code of the process that run
        """
        if self.echo:
            self.print_info(command)

        if return_output:
            # Prepare variable for later assignment
            returncode = None

            # This works but it has no color and no stderr
            def execute(cmd, exit_on_error):
                """Helper function that correctly executes the process and returns
                output.
                """
                popen = subprocess.Popen(
                    cmd,
                    shell=True,
                    bufsize=1,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                    universal_newlines=True,
                )
                for stdout_line in iter(popen.stdout.readline, ""):
                    yield stdout_line
                popen.stdout.close()
                rc = popen.wait()

                # Assign to global, so the value can be seen outside
                nonlocal returncode
                returncode = rc

                if exit_on_error and rc:
                    self.exit(rc)

            output = []

            for line in execute(command, exit_on_error):
                output.append(line)
                if not silent:
                    print(line, end="")

            return {"output": "".join(output), "returncode": returncode}

        else:
            out = None
            err = None

            if silent:
                out = subprocess.DEVNULL
                err = subprocess.STDOUT

            if ignore_sigint:
                previous_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)
                p = subprocess.Popen(command, stdout=out, stderr=err, shell=True)
                p.wait()
                signal.signal(signal.SIGINT, previous_handler)
            else:
                p = subprocess.Popen(command, stdout=out, stderr=err, shell=True)
                p.wait()

            # Should we exit on the error?
            if exit_on_error and p.returncode:
                self.exit(p.returncode)

            return {"output": "", "returncode": p.returncode}

    def run_west(self, west_command: str, **kwargs) -> str:
        """Run wrapper which should be used when executing commands with west tool.

        If toolchain for the detected ncs version is installed then west through
        nrfutil toolchain-manager is used. If it is not installed then west is used
        directly.

        Args:
            west_command (str):     west command to execute
            kwargs:                 Anything that is supported by .run method

        Returns:
            Check .run
        """
        cmd = f"west {west_command}"

        if self.use_toolchain_manager:
            # Run west command as arbitrary command through manager
            return self.run_cmd_in_manager(cmd, **kwargs)
        else:
            return self.run(cmd, **kwargs)

    def enter_manager_shell(self):
        """Enters nrfutil toolchain-manager shell using detected NCS version."""
        cmd = (
            f"{self.consts['nrfutil_path']} toolchain-manager launch --ncs-version"
            f" {self.detected_ncs_version} --shell"
        )

        return self.run(cmd)

    def run_manager(self, command: str, **kwargs):
        """Executes a command with nrfutil toolchain-manager executable.

        This is not suitable to be used with a type of a 'launch -- <command>' command.
        For that run_cmd_in_manager should be used.

        Args:
            command (str):      Command to execute.
            kwargs:             Anything that is supported by .run method.

        Returns:
            Check .run
        """
        cmd = f"{self.consts['nrfutil_path']} toolchain-manager {command}"

        return self.run(cmd, **kwargs)

    def run_cmd_in_manager(self, command: str, exit_on_error: bool = True, **kwargs):
        """Run an arbitrary command through nrfutil toolchain-manager.

        This method should be used when passing any arbitrary command, like west command.

        Args:
            command (str):          Command to execute.
            exit_on_error (bool):   If true the program is exited if the return code of
                                    the ran command is not 0.
            **kwargs:               Anything that is supported by .run method.

        Returns:
            Check .run
        """
        cmd = (
            f"{self.consts['nrfutil_path']} toolchain-manager launch --ncs-version"
            f" {self.detected_ncs_version} -- {command}"
        )

        return self.run(cmd, **kwargs)

    def check_exe(self, exe: str, on_fail_exit: bool = False) -> bool:
        """Checks if the given executable can be found by the which command.

        If on_fail_exit is true it exits the program.

        Args:
            exe (str):              executable to find
            on_fail_exit (bool):    If true it exits cli on exit

        Returns:
            True if given executable was found, false otherwise.
        """
        if not which(exe):
            if on_fail_exit:
                self.exit()
            return False

        return True

    def check_version(self, exe, expected_version, version_cmd="--version"):
        """Checks for version of provided exe program and compares it against
        provided one.

        """
        # WARN: Check version is not yet used anywhere, behaviour yet needs to be
        # verified
        response = self.run(f"{exe} {version_cmd}", silent=True, return_output=True)

        return True if expected_version in response["output"] else False

    def pre_workspace_command_check(
        self,
        ignore_uninstalled_ncs: bool = False,
        ignore_unsupported_ncs: bool = True,
        check_only_west_workspace: bool = False,
    ):
        """This function contains a list of checks that every workspace (not system)
        command should call before executing its actual command.

        In current implementation it does general things:
        * Asserts that workspace command was run from the west workspace
        * Tries to load east.yml
        * Tries to determine the NCS version that is used in the project.
        * Tries to find the version of the toolchain in the nrfutil toolchain-manager.

        This function esentially tries to answer the question: should the underlying
        west command be passed to the nrfutil toolchain-manager or directly to the
        west.

        Args:
            ignore_uninstalled_ncs (bool):  When true, do not exit if detected
                                            NCS version is not installed by the
                                            toolchain-manager. Workspace commands such
                                            as build, flash, clean should set this to
                                            False. install toolchain command should set
                                            this to True.

            ignore_unsupported_ncs (bool):  When true, do not exit if detected
                                            NCS version is not supported by the
                                            toolchain-manager. Workspace commands such
                                            as build, flash, clean should set this to
                                            True. install toolchain command should set
                                            this to False.

            check_only_west_workspace (bool): When true, only check if we are in the
                                              west Workspace, do not check for the rest
                                              of the things. This is useful for the
                                              codechecker command, which should be run
                                              inside the west workspace, but does not
                                              need nrf toolchain manager.
        """
        if self.in_docker:
            # Running in docker, we shouldn't use the toolchain manager.
            self.use_toolchain_manager = False
            return

        # Workspace commands can only be run inside west workspace, so exit if that is
        # not the case.
        if not self.west_dir_path:
            self.print(not_in_west_workspace_msg, highlight=False)
            self.exit()

        if check_only_west_workspace:
            return

        # Exit if east.yml could not be loaded from the project dir; it is not an error
        # if it does not exist, we support that.
        try:
            self.east_yml = load_east_yml(self.project_dir)
        except EastYmlLoadError as msg:
            self.print(format_east_yml_load_error_msg(msg), highlight=False)
            self.exit()

        # Check if ncs version was even detected, this can happen in the cases where
        # normal zephyr repo is used.
        if self.detected_ncs_version is None:
            self.use_toolchain_manager = False
            return

        # Exit if manager is not installed.
        if not self.check_exe(self.consts["nrfutil_path"]):
            self.print(no_toolchain_manager_msg, highlight=False)
            self.exit()

        # If it is installed then we should use it.
        self.use_toolchain_manager = True

        # Early exit if toolchain for the detected ncs version is installed.
        result = self.run_manager("list", silent=True, return_output=True)
        if self.detected_ncs_version in result["output"]:
            self.detected_ncs_version_installed = True
            return

        # Check if toolchain for detected ncs version is supported
        result = self.run_manager("search", silent=True, return_output=True)

        if self.detected_ncs_version in result["output"]:
            # Supported but not installed, should we exit program or silently pass?
            if ignore_uninstalled_ncs:
                return

            self.print(no_toolchain_msg(self), highlight=False)
            self.exit()

        # Not supported, should we silently pass or exit program with message?
        if ignore_unsupported_ncs:
            # Silently pass
            return

        # Exit program, this is usually happens, if we want to install the toolchain.
        self.print(
            ncs_version_not_supported_msg(self, result["output"]), highlight=False
        )
        self.exit()

    def check_for_new_east_version(self):
        """Occasionally check if there is a new version of east available.

        If there is print a message to the user.
        """
        # Check if the file exists
        check_file = os.path.join(self.consts["cache_dir"], "last_version_check")
        if not os.path.isfile(check_file):
            # File does not exist, create it and write current time to it
            with open(check_file, "w") as f:
                f.write(str(time.time()))
            return

        # File exists, read the time from it
        with open(check_file, "r") as f:
            last_check = float(f.read())

        # Check if the time difference is more than 2 hours
        if (time.time() - last_check) < 2 * 3600:
            return

        # More than 2 hours, send a request
        try:
            response = requests.get("https://pypi.org/pypi/east-tool/json", timeout=1)
        except requests.exceptions.RequestException:
            # Something went wrong, do not do anything
            return

        # Check if the request was successful
        if response.status_code != 200:
            # Something went wrong, do not do anything
            return

        # Check if the version is different
        pypi_version = response.json()["info"]["version"]
        current_version = importlib.metadata.version("east-tool")
        if pypi_version != current_version:
            # Print a message that there is a new version available
            msg = (
                "\n[bold yellow]New version of east is available![/]\n"
                "Run [bold]pip install --upgrade east-tool[/] to update.\n"
            )
            self.print(Panel(msg))

        # Write current time to the file to track when the last check was performed
        with open(check_file, "w") as f:
            f.write(str(time.time()))



import argparse
import copy
import os
import shutil as sh

import click
from rich_click.rich_command import RichCommand

from ..east_context import east_command_settings
from ..helper_functions import clean_up_extra_args
from .build_type_flag import construct_extra_cmake_arguments
from .codechecker_helpers import create_codecheckerfile


@click.command(**east_command_settings)
@click.pass_obj
def clean(east):
    """Clean the build folder in current directory.

    \b
    \n\n[bold]Note:[/] This command can be only run from inside of a [bold yellow]West workspace[/].
    """
    east.pre_workspace_command_check()

    east.run("rm -fr build")


class SpecialCommand(RichCommand):
    """Special command that allows to access the raw args."""

    def parse_args(self, ctx, args):
        """Parse the args and store a copy of them in the context."""
        # Before parsing the args, make a copy of them so they can be used later.
        # This is needed because "--" is silently removed from the args by the click,
        # but it is needed to determine the source dir.
        ctx.raw_args = copy.deepcopy(args)
        return super(SpecialCommand, self).parse_args(ctx, args)


@click.command(
    cls=SpecialCommand,
    options_metavar="[options]",
    context_settings=dict(ignore_unknown_options=True, allow_extra_args=True),
)
@click.option(
    "-u",
    "--build-type",
    type=str,
    help=(
        "Which build type (a group of [bold]Kconfig[/] fragment files) to use. Requires"
        " [bold yellow]east.yml[/] with possible build types with specified apps and "
        "samples."
    ),
)
@click.option(
    "--spdx",
    is_flag=True,
    help=(
        "Create an SPDX 2.2 tag-value bill of materials following the completion "
        "of a Zephyr build."
    ),
)
@click.option(
    "--extra-help",
    is_flag=True,
    help="Print help of the [bold magenta]west build[/] command.",
)
@click.argument("args", nargs=-1, type=click.UNPROCESSED, metavar="")
@click.pass_obj
@click.pass_context
def build(ctx, east, build_type, spdx, extra_help, args):
    """Build firmware in the current directory.

    \b
    \n\nInternally runs [magenta bold]west build[/] command, all given arguments are passed directly to it. To learn more about possible [magenta bold]west build[/] arguments and options use --extra-help flag.


    \n\nAfter the build step, if the file [bold]compile_commands.json[/] is found in build dir it will be copied to the project and top west directory. This makes job of locating this file easier for [bold yellow]clangd[/].


    \n\nTo pass additional arguments to the [bold]CMake[/] invocation performed by the [magenta
    bold]west build[/], pass them after a [bold white]"--"[/] at the end of the command line.

    \n\nIf the source dir is listed in the [bold yellow]east.yml[/] then [bold yellow]build type[/] functionality is enabled. East will add a group of [bold]Kconfig[/] fragment files to the build as specified by the selected build type and [bold yellow]east.yml[/] file. See [bold]docs/configuration.md[/] for more info.

    \n\n[bold]Note:[/] When using build types functionality make sure to not pass CONF_FILE and OVERLAY_CONFIG variables to the [bold]CMake[/].
    \n\n[bold]Note:[/] This command can be only run from inside of a [bold yellow]West workspace[/].
    """
    _ = args

    east.pre_workspace_command_check()

    if extra_help:
        east.run_west("build --help")
        east.exit(return_code=0)

    build_cmd, opts = create_build_command_from_commandline(
        east, ctx.raw_args, build_type
    )
    build_dir = opts.build_dir if opts.build_dir else "build"

    if spdx:
        east.run_west(f"spdx --init --build-dir {build_dir}")

    east.run_west(build_cmd)

    if spdx:
        east.run_west(f"spdx --build-dir {build_dir} --analyze-includes --include-sdk")

    compile_file = os.path.join("build", "compile_commands.json")
    if os.path.isfile(compile_file):
        for dest in [east.project_dir, east.west_dir_path]:
            sh.copyfile(compile_file, os.path.join(dest, "compile_commands.json"))

    create_codecheckerfile(
        east, opts.board, build_type, opts.build_dir, opts.source_dir
    )


def create_build_command_from_commandline(east, raw_args, build_type):
    """Helper for creating a build command.

    It parses raw_args and creates two objects:
    - build_cmd: a string with the build command, intended to be given to run_west()
    - opts: an object with parsed arguments
    """
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument("-b", "--board")
    parser.add_argument("-d", "--build-dir")
    parser.add_argument("-t", "--target")

    # find source_dir in the args
    source_dir = None
    option_flag = False

    for arg in raw_args:
        if arg == "--":
            break
        if arg.startswith("-"):
            option_flag = True
            continue
        elif option_flag:
            option_flag = False
            continue
        else:
            source_dir = arg
            break

    # find cmake_args in the args
    cmake_args = raw_args[raw_args.index("--") + 1 :] if "--" in raw_args else None

    # Parse the rest of the args
    opts, _ = parser.parse_known_args(raw_args)
    opts.source_dir = source_dir

    build_cmd = create_build_command(
        east,
        opts.board,
        build_type,
        opts.build_dir,
        opts.target,
        source_dir,
        cmake_args,
    )

    return build_cmd, opts


def create_build_command(
    east,
    board=None,
    build_type=None,
    build_dir=None,
    target=None,
    source_dir=None,
    cmake_args=None,
    silence_diagnostic=False,
):
    """Helper for creating a build command. This extra helper is needed so it can also
    be reused by release command.

    Returns:
        build_cmd: a string with the build command, intended to be given to run_west()
    """
    build_type_args, diagnostic = construct_extra_cmake_arguments(
        east,
        build_type,
        board,
        build_dir,
        source_dir,
    )

    if diagnostic and not silence_diagnostic:
        east.print(diagnostic)

    build_cmd = "build"

    if board:
        build_cmd += f" -b {board}"
    if build_dir:
        build_cmd += f" -d {build_dir}"
    if target:
        build_cmd += f" -t {target}"
    if source_dir:
        build_cmd += f" {source_dir}"

    # Some flags need to be passed as extra parameters to the west tool
    if build_type_args or cmake_args:
        build_cmd += " --"

    if build_type_args:
        build_cmd += f" {build_type_args}"

    if cmake_args:
        build_cmd += f" {clean_up_extra_args(cmake_args)}"

    return build_cmd


@click.command(
    **east_command_settings,
    context_settings=dict(ignore_unknown_options=True, allow_extra_args=True),
)
@click.pass_obj
@click.option(
    "--extra-help",
    is_flag=True,
    help="Print help of the [bold magenta]west flash[/] command.",
)
@click.argument("args", nargs=-1, type=click.UNPROCESSED, metavar="")
def flash(east, extra_help, args):
    """Flash binary to the board's flash.

    \b
    \n\nInternally runs [magenta bold]west flash[/] command, all given arguments are passed directly to it.

    \n\nTo learn more about possible [magenta bold]west flash[/] arguments and options use --extra-help flag.


    \n\n[bold]Note:[/] This command can be only run from inside of a [bold yellow]West workspace[/].
    """
    east.pre_workspace_command_check()

    cmd = "flash "

    if extra_help:
        cmd += "--help"
        east.run_west(cmd)
        east.exit(return_code=0)

    if args:
        cmd += f"{clean_up_extra_args(args)} "

    east.run_west(cmd)


@click.command(**east_command_settings)
@click.option(
    "-s",
    "--shell",
    is_flag=True,
    help=(
        "Launch a sub-shell within the current terminal inside the isolated "
        "environment provided by the [magenta bold]nrfutil toolchain-manager[/]. "
        "Commands after [bold]--[/] are ignored. To exit the sub-shell type "
        "[bold]exit[/] into it and hit ENTER."
    ),
)
@click.argument("args", nargs=-1, type=str, metavar="-- [args]")
@click.pass_obj
def bypass(east, shell, args):
    """Bypass any set of commands directly to the [magenta bold]nrfutil toolchain-manager[/].

    \b
    \n\nPassing any set of commands after double dash [bold]--[/] will pass them directly to
    the [bold magenta]nrfutil toolchain-manager[/] executable.

    \n\nThose commands will run in the context of the isolated environment, which is provided by the executable.

    \n\nExample:

    \n\nCommand [bold]east bypass -- west build -b nrf52840dk_nrf52840[/]
    \n\nbecomes [bold]west build -b nrf52840dk_nrf52840[/]



    \n\n[bold]Note:[/] This command can be only run from inside of a [bold yellow]West workspace[/].
    """
    east.pre_workspace_command_check()

    if not east.detected_ncs_version:
        east.print(
            "No version of [bold cyan]nRF Connect SDK[/] was detected in this "
            "[bold yellow]West workspace[/], can't run [bold magenta]east bypass[/]."
        )
        east.exit()

    if shell:
        east.enter_manager_shell()
        east.exit(return_code=0)

    if not args:
        east.exit()

    if not east.use_toolchain_manager:
        east.exit(
            "nrfutil toolchain manager is not available in this [bold yellow]West workspace[/]."
        )

    cmd = clean_up_extra_args(args)
    east.run_cmd_in_manager(cmd)


@click.command(
    **east_command_settings,
    context_settings=dict(ignore_unknown_options=True, allow_extra_args=True),
)
@click.pass_obj
@click.option(
    "--extra-help",
    is_flag=True,
    help="Print help of the [bold magenta]west debug[/] command.",
)
@click.argument("args", nargs=-1, type=click.UNPROCESSED, metavar="")
def debug(east, extra_help, args):
    """Connect to the board, flash the program, and start a debugging session.

    \b
    \n\nInternally runs [magenta bold]west debug[/] command, all given arguments are passed directly to it.

    \n\nTo learn more about possible [magenta bold]west debug[/] arguments and options use --extra-help flag.

    \n\n[bold]Note:[/] Add --tui flag to use Text User Interface.
    \n\n[bold]Note:[/] This command can be only run from inside of a [bold yellow]West workspace[/].
    """
    east.pre_workspace_command_check()

    if extra_help:
        east.run_west("debug --help")
    else:
        east.run_west("debug " + clean_up_extra_args(args), ignore_sigint=True)


@click.command(
    **east_command_settings,
    context_settings=dict(ignore_unknown_options=True, allow_extra_args=True),
)
@click.pass_obj
@click.option(
    "--extra-help",
    is_flag=True,
    help="Print help of the [bold magenta]west attach[/] command.",
)
@click.argument("args", nargs=-1, type=click.UNPROCESSED, metavar="")
def attach(east, extra_help, args):
    """Like "east debug", but doesn't reflash the program.

    \b
    \n\nInternally runs [magenta bold]west attach[/] command, all given arguments are passed directly to it.

    \n\nTo learn more about possible [magenta bold]west attach[/] arguments and options use --extra-help flag.

    \n\n[bold]Note:[/] Add --tui flag to use Text User Interface.
    \n\n[bold]Note:[/] This command can be only run from inside of a [bold yellow]West workspace[/].
    """
    east.pre_workspace_command_check()

    if extra_help:
        east.run_west("attach --help")
    else:
        east.run_west("attach " + clean_up_extra_args(args), ignore_sigint=True)


@click.command(
    **east_command_settings,
    context_settings=dict(ignore_unknown_options=True, allow_extra_args=True),
)
@click.pass_obj
@click.option(
    "--extra-help",
    is_flag=True,
    help="Print help of the [bold magenta]west twister[/] command.",
)
@click.argument("args", nargs=-1, type=click.UNPROCESSED, metavar="")
def twister(east, extra_help, args):
    """Run Twister, a test runner tool.

    \b
    \n\nInternally runs [magenta bold]west twister[/] command, all given arguments are passed directly to it.

    \n\nTo learn more about possible Twister arguments and options use --extra-help flag.


    \n\n[bold]Note:[/] This command can be only run from inside of a [bold yellow]West workspace[/].
    """
    east.pre_workspace_command_check()

    cmd = "twister "

    if extra_help:
        cmd += "--help"
        east.run_west(cmd)
        east.exit(return_code=0)

    if args:
        cmd += f"{clean_up_extra_args(args)} "

    east.run_west(cmd)



import os

import pykwalify.core
import pykwalify.errors
import yaml

from .helper_functions import return_dict_on_match


class EastYmlLoadError(RuntimeError):
    """Some error happened when trying to load east.yml."""


def format_east_yml_load_error_msg(exception_msg):
    """Format error message for EastYmlLoadError.

    Use this to format error messages that happen when trying to load east.yml
    """
    return (
        "An [bold red]error[/] occurred when trying to load [bold yellow]east.yml[/]"
        f" file!\n\n[italic yellow]{exception_msg}[/]\n"
    )


def load_east_yml(project_dir: str):
    """Try to load east.yml. If that succeeds validate it.

        project_dir (str): Path to project directory, where east.yml should be located.

    Returns:
        dict with east.yml contents if east.yml is found and correctly validated. If it
        can not be found it returns None.

    """
    east_yml = os.path.join(project_dir, "east.yml")

    if not os.path.isfile(east_yml):
        return None

    # Validate yaml
    schema_yml = os.path.join(os.path.dirname(__file__), "configuration-schema.yaml")
    try:
        c = pykwalify.core.Core(source_file=east_yml, schema_files=[schema_yml])
    except pykwalify.errors.CoreError:
        # This error is raised when east.yml is empty, which is allowed.
        return None

    try:
        c.validate(raise_exception=True)
    except pykwalify.core.SchemaError as e:
        raise EastYmlLoadError(e)

    # Load file
    with open(east_yml, "r") as file:
        east_yml = yaml.safe_load(file)

    # Handle duplicated entries in app name, samples names and build types under each
    # app
    def check_duplicated_entries(array_of_dicts, printed_key, subkey):
        names = [ele[subkey] for ele in array_of_dicts]
        if len(names) > len(set(names)):
            raise EastYmlLoadError(
                f"Found duplicated [bold]{subkey}s[/] under [bold]{printed_key}[/] key"
                " in [bold]east.yml[/]!"
            )

    # This is needed to discern between apps key present, apps key present, but not set
    # and not app key at all.
    apps = east_yml.get("apps", -1)

    # apps are optional
    if apps is not None and apps != -1:
        check_duplicated_entries(east_yml["apps"], "apps", "name")
        for app in east_yml["apps"]:
            if "build-types" in app:
                check_duplicated_entries(
                    app["build-types"], f"{app['name']}.build-types", "type"
                )

    if apps is None:
        # Exists, but it was not set
        raise EastYmlLoadError(
            "[bold]apps[/] key in [bold yellow]east.yml[/] has no apps listed under it!"
        )

    # samples are optional
    if east_yml.get("samples"):
        check_duplicated_entries(east_yml["samples"], "samples", "name")
        # For each sample that has inherit key check if that app with that build type
        # exists
        for sample in east_yml["samples"]:
            if "inherit-build-type" in sample:
                # Inherit needs apps
                if not east_yml.get("apps"):
                    raise EastYmlLoadError(
                        f"Sample [bold]{sample['name']}[/] is trying to inherit, but"
                        " there are no apps to inherit from!"
                    )

                inherited_app = sample["inherit-build-type"]["app"]
                inherited_type = sample["inherit-build-type"]["build-type"]

                app = return_dict_on_match(east_yml["apps"], "name", inherited_app)
                if not app:
                    raise EastYmlLoadError(
                        f"Sample [bold]{sample['name']}[/] is trying to inherit from a"
                        f" [bold red]non-existing[/] app [bold]{inherited_app}[/]."
                    )
                # Release build type is special, apps always have it implicitly, so we
                # do not have to search for it.
                if inherited_type != "release":
                    if "build-types" not in app or not return_dict_on_match(
                        app["build-types"], "type", inherited_type
                    ):
                        raise EastYmlLoadError(
                            f"Sample [bold]{sample['name']}[/] is trying to inherit "
                            "from a [bold red]non-existing[/] build-type"
                            f" [bold]{inherited_type}[/]."
                        )

    return east_yml



import os
import shutil

from click.testing import CliRunner

from east.__main__ import cli
from east.east_context import EastContext

west_config_content = """
[manifest]
path = project
file = west.yml

[zephyr]
base = zephyr
"""
nrfsdk_yaml_content = """
manifest:
  self:
    # This repository should be cloned to
    path: project

  remotes:
    - name: nrfconnect
      url-base: https://github.com/nrfconnect

  projects:
    # the NCS repository
    - name: nrf
      repo-path: sdk-nrf
      remote: nrfconnect
      revision: v2.1.0
      import: true

    - name: zephyr
      remote: nrfconnect
      repo-path: sdk-zephyr
      revision: v3.1.99-ncs1
      import: true
"""

zephyrsdk_yaml_content = """
manifest:
  self:
    path: example-application
    west-commands: scripts/west-commands.yml

  remotes:
    - name: zephyrproject-rtos
      url-base: https://github.com/zephyrproject-rtos

  projects:
    - name: zephyr
      remote: zephyrproject-rtos
      revision: main
      import: true
"""

east_yaml_single_app = """
apps:
  - name: test_one
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840

    build-types:
      - type: debug
        conf-files:
          - debug.conf
      - type: uart
        conf-files:
          - debug.conf
          - uart.conf

samples:
  - name: settings
    west-boards:
      - custom_nrf52840dk
    inherit-build-type:
        app: test_one
        build-type: debug

  - name: dfu
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840
    # Don't inherit, use prj.conf in the sample's folder.
"""

east_yaml_multiple_apps = """
apps:
  - name: test_one
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840

    build-types:
      - type: debug
        conf-files:
          - debug.conf
      - type: uart
        conf-files:
          - debug.conf
          - uart.conf

  - name: test_two
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840

    build-types:
      - type: debug
        conf-files:
          - debug.conf
      - type: rtt
        conf-files:
          - rtt.conf
      - type: debug-rtt
        conf-files:
          - debug.conf
          - rtt.conf

samples:
  - name: settings
    west-boards:
      - custom_nrf52840dk
    inherit-build-type:
        app: test_one
        build-type: debug

  - name: dfu
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840
    # Don't inherit, use prj.conf in the sample's folder.
"""


def create_and_write(path: str, filename: str, content: str = None):
    """Create a file in the given path with the given content.

    Args:
        path (str):         Path from where to create the filename
        filename (str):     Filename can be either a direct file name such as a file.txt
                            or a longer path, such as dir_a/dir_b/file.txt
        content (str):      File content to write. If None, nothing is written.

    """
    filepath = os.path.join(path, filename)

    os.makedirs(os.path.dirname(filepath), exist_ok=True)

    with open(filepath, "w", encoding="utf-8") as f:
        if content:
            f.write(content)


dummy_config = """
CONFIG_DEBUG=y
"""
dummy_cmakelists_txt = """
cmake_minimum_required(VERSION 3.20.0)

find_package(Zephyr REQUIRED HINTS $ENV{ZEPHYR_BASE})
project(test_project)

target_sources(app PRIVATE src/main.c)
"""


def _create_good_west_workspace(west_top_dir):
    """Main function, which will create correct west workspace. All other 'bad'
    functions will just delete from it.
    """
    folders = [
        os.path.join("project", "app"),
        os.path.join("project", "samples", "settings"),
        os.path.join("project", "samples", "dfu"),
        os.path.join("project", "tests", "basic"),
        "zephyr",
    ]

    for folder in folders:
        os.makedirs(os.path.join(west_top_dir, folder), exist_ok=True)

    create_and_write(west_top_dir, ".west/config", west_config_content)
    create_and_write(west_top_dir, "project/west.yml", nrfsdk_yaml_content)
    create_and_write(west_top_dir, "project/east.yml", east_yaml_single_app)
    create_and_write(
        west_top_dir, "project/app/conf/nrf52840dk_nrf52840.conf", dummy_config
    )
    create_and_write(west_top_dir, "project/app/CMakeLists.txt", dummy_cmakelists_txt)

    # Board files, just create them, you do not need to write anything.
    board_path = "project/boards/arm/custom_nrf52840dk"
    board_files = [
        "board.cmake",
        "Kconfig",
        "Kconfig.board",
        "Kconfig.defconfig",
        "revision.cmake",
        "custom_nrf52840dk - pinctrl.dtsi",
        "custom_nrf52840dk.dts",
        "custom_nrf52840dk.yaml",
        "custom_nrf52840dk_1_0_0.conf",
        "custom_nrf52840dk_1_0_0.overlay",
        "custom_nrf52840dk_1_1_0.conf",
        "custom_nrf52840dk_1_1_0.overlay",
        "custom_nrf52840dk_2_20_1.conf",
        "custom_nrf52840dk_2_20_1.overlay",
        "custom_nrf52840dk_defconfig",
    ]
    for board_file in board_files:
        create_and_write(west_top_dir, os.path.join(board_path, board_file))
    board_path = "project/boards/arm/nrf52840dk_nrf52840"
    board_files = [
        "board.cmake",
        "Kconfig",
        "Kconfig.board",
        "Kconfig.defconfig",
        "revision.cmake",
        "nrf52840dk_nrf52840.dts",
        "nrf52840dk_nrf52840.yaml",
        "nrf52840dk_nrf52840_defconfig",
    ]
    for board_file in board_files:
        create_and_write(west_top_dir, os.path.join(board_path, board_file))

    return os.path.join(west_top_dir, "project")


def create_good_west(west_top_dir):
    """Creates a correct west workspace in west_top_dir.

    Args:
        west_top_dir (str): Path to west top dir.

    Returns:
        Path to the project inside west top dir.
    """
    return _create_good_west_workspace(west_top_dir)


def create_good_west_multi_app(west_top_dir):
    """Creates a correct west workspace in west_top_dir, with multi app east yaml.

    Args:
        west_top_dir (str): Path to west top dir.

    Returns:
        Path to the project inside west top dir.
    """
    create_and_write(west_top_dir, "project/east.yml", east_yaml_multiple_apps)

    # We need this to clean any files left by create_good_west
    shutil.rmtree(os.path.join(west_top_dir, "project/app"))
    os.makedirs(os.path.join(west_top_dir, "project/app/test_one"), exist_ok=True)
    os.makedirs(os.path.join(west_top_dir, "project/app/test_two"), exist_ok=True)
    create_and_write(
        west_top_dir, "project/app/test_one/conf/nrf52840dk_nrf52840.conf", dummy_config
    )


def west_no_nrf_sdk_in_yaml(west_top_dir):
    """Creates a west workspace without nrf-sdk in west.yaml.

    Args:
        west_top_dir (): Path to west top dir.
    """
    create_and_write(west_top_dir, ".west/config", west_config_content)
    create_and_write(west_top_dir, "project/west.yml", zephyrsdk_yaml_content)


def assert_strings_equal(string1: str, string2: str):
    """Helper that should be used when comparing strings that come from east's stdout
    and east's internal hardcoded strings.

    Returns:
        Asserts if strings are different
    """

    def clear_rich(string):
        """Output from runner.invoke and hard-coded messages can contain different
        number of newlines, and indent characters, this is preventing comparisons in
        asserts.
        """
        return string.replace("\n", "").replace("\t", 8 * " ")

    assert clear_rich(string1) == clear_rich(string2)


def helper_test_against_west_run(
    monkeypatch,
    mocker,
    path,
    east_cmd,
    expected_west_cmd=None,
    expected_west_cmds=None,
    should_succed=True,
):
    """Helper function for making tests easier to read.

    Args:
        monkeypatch ():         fixture
        mocker ():              fixture
        path ():                To which path should we change
        east_cmd ():            which east command should be called
        expected_west_cmd ():   A single expected west cmd, a string.
        expected_west_cmds ():  A List of expected west_cmds.
        should_succed ():       If true then the command should succeded.

    Only a expected_west_cmd or expected_west_cmds can be given, both not both.
    If none is given then no run_west call should happend.

    Returns:
        Result object, which can be further checked.
    """
    runner = CliRunner()

    # Mock output of git commmand, so tests do not have to depend on it
    mocker.patch(
        "east.workspace_commands.release_commands.get_git_version",
        return_value={"tag": "v1.0.0.", "hash": ""},
    )
    mocker.patch(
        "east.workspace_commands.codechecker_helpers.get_git_version",
        return_value={"tag": "v1.0.0.", "hash": ""},
    )

    monkeypatch.chdir(path)
    mocker.patch(
        "east.east_context.EastContext.run_west",
        return_value={"output": "", "returncode": 0},
    )

    # Setting catch_exceptions to False enables us to see programming errors in East
    # code
    result = runner.invoke(cli, east_cmd.strip().split(" "), catch_exceptions=False)

    run_west = EastContext.run_west

    if expected_west_cmds:
        # This conversion is needed due to assert_has_calls interface
        calls = [
            mocker.call(cmd, exit_on_error=False, silent=True, return_output=True)
            for cmd in expected_west_cmds
        ]
        run_west.assert_has_calls(calls)

    elif expected_west_cmd:
        run_west.assert_called_once_with(expected_west_cmd)
    else:
        run_west.assert_not_called()

    expected_return_code = 0 if should_succed else 1

    assert result.exit_code == expected_return_code
    return result


def helper_test_against_west_dbg(
    monkeypatch,
    mocker,
    path,
    east_cmd,
    expected_west_cmd=None,
    expected_west_cmds=None,
    should_succed=True,
):
    """Helper function for tests easier to debug.

    Running this instead of the real helper will just make sure that the correct command
    runs, but it will not execute any east.run_west() calls.

    Args:
        monkeypatch ():         fixture
        mocker ():              fixture
        path ():                To which path should we change
        east_cmd ():            which east command should be called
        expected_west_cmd ():   Not used
        expected_west_cmds ():  Not used
        should_succed ():       Not used
    """
    _ = expected_west_cmd
    _ = expected_west_cmds
    _ = should_succed

    runner = CliRunner()

    # Mock output of git commmand, so tests do not have to depend on it
    mocker.patch(
        "east.workspace_commands.release_commands.get_git_version",
        return_value={"tag": "v1.0.0.", "hash": ""},
    )

    monkeypatch.chdir(path)
    mocker.patch("east.east_context.EastContext.run_west")

    result = runner.invoke(cli, east_cmd.strip().split(" "))
    print(result.output)
